from __future__ import print_function, division, absolute_import

from ..utils import ignoring

from .cluster import Cluster
from .local import LocalCluster
from .spec import SpecCluster
from .adaptive import Adaptive

with ignoring(ImportError):
    from .ssh import SSHCluster
